// Copyright (c) 2024 Celestino Amoroso (celestino.amoroso@gmail.com).
// All rights reserved.

// funcs-math.go
package expr

import (
	"fmt"
	"io"
)

func checkNumberParamExpected(funcName string, paramValue any, paramPos int) (err error) {
	if !(isNumber(paramValue) || isList(paramValue) || isFraction(paramValue)) {
		err = fmt.Errorf("%s(): param nr %d has wrong type %T, number expected", funcName, paramPos+1, paramValue)
	}
	return
}

func doAdd(ctx ExprContext, name string, it Iterator) (result any, err error) {
	var sumAsFloat, sumAsFract bool
	var floatSum float64 = 0.0
	var intSum int64 = 0
	var fractSum *fraction
	var v any

	for v, err = it.Next(); err == nil; v, err = it.Next() {
		if subIter, ok := v.(Iterator); ok {
			if v, err = doAdd(ctx, name, subIter); err != nil {
				break
			}
			if subIter.HasOperation(cleanName) {
				if _, err = subIter.CallOperation(cleanName, nil); err != nil {
					return
				}
			}
		} else {
			if err = checkNumberParamExpected(name, v, it.Index()); err != nil {
				break
			}
			if list, ok := v.(*ListType); ok {
				if v, err = doAdd(ctx, name, NewListIterator(list)); err != nil {
					break
				}
			}
		}

		if !sumAsFloat {
			if isFloat(v) {
				sumAsFloat = true
				if sumAsFract {
					floatSum = fractSum.toFloat()
				} else {
					floatSum = float64(intSum)
				}
			} else if !sumAsFract && isFraction(v) {
				fractSum = newFraction(intSum, 1)
				sumAsFract = true
			}
		}

		if sumAsFloat {
			floatSum += numAsFloat(v)
		} else if sumAsFract {
			var item *fraction
			var ok bool
			if item, ok = v.(*fraction); !ok {
				iv, _ := v.(int64)
				item = newFraction(iv, 1)
			}
			fractSum = sumFract(fractSum, item)
		} else {
			iv, _ := v.(int64)
			intSum += iv
		}
	}
	if err == nil || err == io.EOF {
		err = nil
		if sumAsFloat {
			result = floatSum
		} else if sumAsFract {
			result = fractSum
		} else {
			result = intSum
		}
	}
	return
}

func addFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	result, err = doAdd(ctx, name, NewArrayIterator(args))
	return
}

func doMul(ctx ExprContext, name string, it Iterator) (result any, err error) {
	var mulAsFloat, mulAsFract bool
	var floatProd float64 = 1.0
	var intProd int64 = 1
	var fractProd *fraction
	var v any

	for v, err = it.Next(); err == nil; v, err = it.Next() {
		if subIter, ok := v.(Iterator); ok {
			if v, err = doMul(ctx, name, subIter); err != nil {
				break
			}
			if subIter.HasOperation(cleanName) {
				if _, err = subIter.CallOperation(cleanName, nil); err != nil {
					return
				}
			}
		} else {
			if err = checkNumberParamExpected(name, v, it.Index()); err != nil {
				break
			}

			if list, ok := v.(*ListType); ok {
				if v, err = doMul(ctx, name, NewListIterator(list)); err != nil {
					break
				}
			}
		}

		if !mulAsFloat {
			if isFloat(v) {
				mulAsFloat = true
				if mulAsFract {
					floatProd = fractProd.toFloat()
				} else {
					floatProd = float64(intProd)
				}
			} else if !mulAsFract && isFraction(v) {
				fractProd = newFraction(intProd, 1)
				mulAsFract = true
			}
		}

		if mulAsFloat {
			floatProd *= numAsFloat(v)
		} else if mulAsFract {
			var item *fraction
			var ok bool
			if item, ok = v.(*fraction); !ok {
				iv, _ := v.(int64)
				item = newFraction(iv, 1)
			}
			fractProd = mulFract(fractProd, item)
		} else {
			iv, _ := v.(int64)
			intProd *= iv
		}
	}
	if err == nil || err == io.EOF {
		err = nil
		if mulAsFloat {
			result = floatProd
		} else if mulAsFract {
			result = fractProd
		} else {
			result = intProd
		}
	}
	return
}

func mulFunc(ctx ExprContext, name string, args []any) (result any, err error) {
	result, err = doMul(ctx, name, NewArrayIterator(args))
	return
}

func ImportMathFuncs(ctx ExprContext) {
	ctx.RegisterFunc("add", &simpleFunctor{f: addFunc}, 0, -1)
	ctx.RegisterFunc("mul", &simpleFunctor{f: mulFunc}, 0, -1)
}

func init() {
	registerImport("math.arith", ImportMathFuncs, "Function add() and mul()")
}
